/*
  +----------------------------------------------------------------------+
  | PHP Version 5                                                        |
  +----------------------------------------------------------------------+
  | Copyright (c) 1997-2004 The PHP Group                                |
  +----------------------------------------------------------------------+
  | This source file is subject to version 3.0 of the PHP license,       |
  | that is bundled with this package in the file LICENSE, and is        |
  | available through the world-wide-web at the following url:           |
  | http://www.php.net/license/3_0.txt.                                  |
  | If you did not receive a copy of the PHP license and are unable to   |
  | obtain it through the world-wide-web, please send a note to          |
  | license@php.net so we can mail you a copy immediately.               |
  +----------------------------------------------------------------------+
  | Author: Oliver Welter <areaz2@php.net>                               |
  +----------------------------------------------------------------------+
*/

/* $Id: header,v 1.15 2004/01/08 16:46:52 sniper Exp $ */

#ifdef PHP_WIN32
#error "This extension is only for Linux based systems"
#endif

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "php.h"
#include "php_ini.h"
#include "ext/standard/info.h"
#include "php_coin_acceptor.h"

#include <sys/time.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/ioctl.h>

#include <linux/fs.h>
#include <linux/kernel.h>

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#include <stdlib.h>
#include <ctype.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <errno.h>
#include <string.h>
#include <termios.h>
#include <stdio.h>

/* e.g. IRIX does not have CRTSCTS */
#ifndef CRTSCTS
# ifdef CNEW_RTSCTS
#  define CRTSCTS CNEW_RTSCTS
# else
#  define CRTSCTS 0
# endif /* CNEW_RTSCTS */
#endif /* !CRTSCTS */

#define le_coin_acceptor_name "Coin Acceptor File Descriptor"

#define CA_ANSWER_LENGTH 4
#define CA_REQUEST_LENGTH 3
#define CA_RESET_LENGTH 4

static char CA_PIT_A[3] = "A\0A";
static char CA_PIT_B[3] = "B\0B";
static char CA_PIT_C[3] = "C\0C";
static char CA_PIT_D[3] = "D\0D";

static char CA_RESET[3] = "R\0R";
static char CA_FLUSH[3] = "F\0F";

static char CA_PIT_A_OK[2] = "A\1";
static char CA_PIT_B_OK[2] = "B\1";
static char CA_PIT_C_OK[2] = "C\1";
static char CA_PIT_D_OK[2] = "D\1";

static char CA_RESET_DONE[3] = "R\1";
static char CA_FLUSH_DONE[3] = "F\1";

static char CA_PIT_A_NULL[2] = "\0B";
static char CA_PIT_B_NULL[2] = "\0C";
static char CA_PIT_C_NULL[2] = "\0D";
static char CA_PIT_D_NULL[2] = "\0E";

#define CA_SLOT_A 0
#define CA_SLOT_B 1
#define CA_SLOT_C 2
#define CA_SLOT_D 3

ZEND_DECLARE_MODULE_GLOBALS(coin_acceptor)

/* True global resources - no need for thread safety here */
static int le_coin_acceptor;

/* {{{ coin_acceptor_functions[]
 *
 * Every user visible function must have an entry in coin_acceptor_functions[].
 */
function_entry coin_acceptor_functions[] = {
	PHP_FE(confirm_coin_acceptor_compiled,	NULL)		/* For testing, remove later. */
	PHP_FE(ca_open, NULL)
	PHP_FE(ca_init, NULL)
	PHP_FE(ca_read, NULL)
	PHP_FE(ca_reset, NULL)
	PHP_FE(ca_close, NULL)
	{NULL, NULL, NULL}	/* Must be the last line in coin_acceptor_functions[] */
};
/* }}} */

/* {{{ coin_acceptor_module_entry
 */
zend_module_entry coin_acceptor_module_entry = {
#if ZEND_MODULE_API_NO >= 20010901
	STANDARD_MODULE_HEADER,
#endif
	"coin_acceptor",
	coin_acceptor_functions,
	PHP_MINIT(coin_acceptor),
	PHP_MSHUTDOWN(coin_acceptor),
	NULL,
	NULL,
	PHP_MINFO(coin_acceptor),
#if ZEND_MODULE_API_NO >= 20010901
	"0.1", 
#endif
	STANDARD_MODULE_PROPERTIES
};
/* }}} */

#ifdef COMPILE_DL_COIN_ACCEPTOR
ZEND_GET_MODULE(coin_acceptor)
#endif

static void _ca_close_fd(zend_rsrc_list_entry *rsrc TSRMLS_DC)
{
        php_fd_t *f = (php_fd_t *) rsrc->ptr;
        close(f->fd);
        free(f);
}

#define RCAC(c) REGISTER_LONG_CONSTANT(#c, c, CONST_CS | CONST_PERSISTENT)
#define CA_UNDEF_CONST -1

/* {{{ PHP_INI
 */
PHP_INI_BEGIN()
    STD_PHP_INI_ENTRY("coin_acceptor.device", 		"/dev/ttyS1", PHP_INI_ALL, OnUpdateString, device, zend_coin_acceptor_globals, coin_acceptor_globals)
    STD_PHP_INI_ENTRY("coin_acceptor.autoreset",	"On", PHP_INI_ALL, OnUpdateString, autoreset, zend_coin_acceptor_globals, coin_acceptor_globals)
    STD_PHP_INI_ENTRY("coin_acceptor.delay",		"53132", PHP_INI_ALL, OnUpdateLong, delay, zend_coin_acceptor_globals, coin_acceptor_globals)
PHP_INI_END()
/* }}} */

/* {{{ php_coin_acceptor_init_globals
 */
static void php_coin_acceptor_init_globals(zend_coin_acceptor_globals *coin_acceptor_globals)
{
	coin_acceptor_globals->device = NULL;
	coin_acceptor_globals->autoreset = NULL;
	coin_acceptor_globals->delay = 0;
}
/* }}} */

/* {{{ PHP_MINIT_FUNCTION
 */
PHP_MINIT_FUNCTION(coin_acceptor)
{
        le_coin_acceptor = zend_register_list_destructors_ex(_ca_close_fd, NULL, le_coin_acceptor_name, module_number);

	RCAC(CA_ANSWER_LENGTH);
	RCAC(CA_REQUEST_LENGTH);
	RCAC(CA_RESET_LENGTH);
	RCAC(CA_SLOT_A);
	RCAC(CA_SLOT_B);
	RCAC(CA_SLOT_C);
	RCAC(CA_SLOT_D);

	ZEND_INIT_MODULE_GLOBALS(coin_acceptor, php_coin_acceptor_init_globals, NULL);
	REGISTER_INI_ENTRIES();

	return SUCCESS;
}
/* }}} */

/* {{{ PHP_MSHUTDOWN_FUNCTION
 */
PHP_MSHUTDOWN_FUNCTION(coin_acceptor)
{
	UNREGISTER_INI_ENTRIES();

	return SUCCESS;
}
/* }}} */

/* {{{ PHP_MINFO_FUNCTION
 */
PHP_MINFO_FUNCTION(coin_acceptor)
{
	php_info_print_table_start();
	php_info_print_table_header(2, "coin_acceptor support", "enabled");
	php_info_print_table_end();

	DISPLAY_INI_ENTRIES();
}
/* }}} */

static void new_php_ca_fd(php_fd_t **f, int fd)
{
        *f = malloc(sizeof(php_fd_t));
        (*f)->fd = fd;
}

static void ca_sleep()
{
        /*if (COIN_ACCEPTOR_G(delay) > 0)
        {
          usleep(COIN_ACCEPTOR_G(delay));
        }
        */
        
        // CURRENTLY STATIC
        usleep(53132);
}

int answer_to_lower(char *string, char *destination)
{
    int length = strlen(string);
    int temp;
    int tmp;
    int i;

    for (i = 0; i < length; i++) {
      temp = (int)string[i];
      tmp = tolower(temp);
      destination[i] = (char)tmp;
    }

    return 0;
}

int ca_reset_slot(int fd, int slot)
{
        char     answer[CA_ANSWER_LENGTH];
        char	 resetreq[CA_ANSWER_LENGTH];
	char	 *pit[4];
	int	 result;
	int	 i;

	pit[CA_SLOT_A] = "A";
	pit[CA_SLOT_B] = "B";
	pit[CA_SLOT_C] = "C";
	pit[CA_SLOT_D] = "D";

	        switch (slot)
	        {
	          case CA_SLOT_A:
	            write(fd, CA_PIT_A, CA_REQUEST_LENGTH);
	            break;
                  case CA_SLOT_B:
                    write(fd, CA_PIT_B, CA_REQUEST_LENGTH);
                    break;
                  case CA_SLOT_C:
                    write(fd, CA_PIT_C, CA_REQUEST_LENGTH);
                    break;
                  case CA_SLOT_D:
                    write(fd, CA_PIT_D, CA_REQUEST_LENGTH);
                    break;
	        }
	        
	        ca_sleep();
	        read(fd, answer, CA_ANSWER_LENGTH);
	        
	        switch (slot)
	        {
	          case CA_SLOT_A:
	            result = strncmp(answer, CA_PIT_A_OK, 1);
	            break;
                  case CA_SLOT_B:
                    result = strncmp(answer, CA_PIT_B_OK, 1);
                    break;
                  case CA_SLOT_C:
                    result = strncmp(answer, CA_PIT_C_OK, 1);
                    break;
                  case CA_SLOT_D:
                    result = strncmp(answer, CA_PIT_D_OK, 1);
                    break;
	        }
	        
	        if (result != 0)
	        {
	          return 1;
	        }

        answer_to_lower(answer, resetreq);
  
        write(fd, resetreq, CA_ANSWER_LENGTH);

        ca_sleep();
        read(fd, answer, CA_ANSWER_LENGTH);

	if (strncmp(answer, resetreq, 1) == 0)
	{
	  return 0;
	} else {
	  return 4;
	}
}

/* Remove the following function when you have succesfully modified config.m4
   so that your module can be compiled into PHP, it exists only for testing
   purposes. */

/* Every user-visible function in PHP should document itself in the source */
/* {{{ proto string confirm_coin_acceptor_compiled(string arg)
   Return a string to confirm that the module is compiled in */
PHP_FUNCTION(confirm_coin_acceptor_compiled)
{
	char *arg = NULL;
	int arg_len, len;
	char string[256];

	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "s", &arg, &arg_len) == FAILURE) {
		return;
	}

	len = sprintf(string, "Congratulations! You have successfully modified ext/%.78s/config.m4. Module %.78s is now compiled into PHP.", "coin_acceptor", arg);
	RETURN_STRINGL(string, len, 1);
}
/* }}} */
/* The previous line is meant for vim and emacs, so it can correctly fold and 
   unfold functions in source code. See the corresponding marks just before 
   function definition, where the functions purpose is also documented. Please 
   follow this convention for the convenience of others editing your code.
*/

/* {{{ proto resource ca_open([string device])
   Open a coin acceptor device */
PHP_FUNCTION(ca_open)
{
        php_fd_t *f;
        char     *ca_device;
        int       ca_device_length;
        int       fd;

        if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "|s", &ca_device, &ca_device_length) == FAILURE) {
                // Try to use the defaults
                ca_device = COIN_ACCEPTOR_G(device);
                ca_device_length = strlen(ca_device);
        }

        if (ca_device) {
                fd = open(ca_device, O_RDWR);
        } else {
                php_error_docref(NULL TSRMLS_CC, E_WARNING, "ca_open() expects parameter 1 to be the device name");
        }

        if (fd == -1) {
                php_error_docref(NULL TSRMLS_CC, E_WARNING, "cannot open device %s: %s", ca_device, strerror(errno));
                RETURN_FALSE;
        }

        new_php_ca_fd(&f, fd);
        ZEND_REGISTER_RESOURCE(return_value, f, le_coin_acceptor);
}

/* {{{ proto bool ca_init(resource fd)
   Initialize a coin acceptor by fd descriptor */
PHP_FUNCTION(ca_init)
{
        zval     *r_fd;
        php_fd_t *f;
        char     *answer;
	int	 result;
	int	 i;
        ssize_t   res;

        if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "r", &r_fd) == FAILURE) {
                return;
        }

        ZEND_FETCH_RESOURCE(f, php_fd_t *, &r_fd, -1, le_coin_acceptor_name, le_coin_acceptor);

	for (i = 0; i < 4; i++)
	{
	        switch (i)
	        {
	          case 0:
	          case 1:
  	            write(f->fd, CA_PIT_A, CA_REQUEST_LENGTH);
  	            break;
	          case 2:
  	            write(f->fd, CA_RESET, CA_REQUEST_LENGTH);
  	            break;
	          case 3:
  	            write(f->fd, CA_FLUSH, CA_REQUEST_LENGTH);
  	            break;
	        }
	        
	        ca_sleep();
	        answer = emalloc(CA_ANSWER_LENGTH + 1);
	        res = read(f->fd, answer, CA_ANSWER_LENGTH);

	        if (res <= 0)
	        {
		  efree(answer);
                  php_error_docref(NULL TSRMLS_CC, E_WARNING, "no answer from device");
                  RETURN_FALSE;       
		}

		answer = erealloc(answer, res + 1);
		answer[res] = 0;
	        
	        switch (i)
	        {
	          case 0:
                  case 1:
                    result = strncmp(answer, CA_PIT_A_OK, 1);
                    break;
                  case 2:
                    result = strncmp(answer, CA_RESET_DONE, 1);
                    break;
                  case 3:
                    result = strncmp(answer, CA_FLUSH_DONE, 1);
                    break;
	        }
	        
	        if (result != 0)
	        {
	            php_error_docref(NULL TSRMLS_CC, E_WARNING, "device has not answered as expected");
	            RETURN_FALSE;       
	        }
	        
	        efree(answer);
	}

	// No need to check the answers
	RETURN_TRUE;	
}

#define ADD_LFIELD(f, v) add_assoc_long_ex(return_value, (f), sizeof(f), v);

/* {{{ proto array ca_read(resource fd)
   Read from coin acceptor by fd descriptor */
PHP_FUNCTION(ca_read)
{
        zval     *r_fd;
        php_fd_t *f;
        char     *answer[CA_ANSWER_LENGTH];
	char	 *pit[4];
	int	 result;
	int	 i;
        ssize_t   res;

        if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "r", &r_fd) == FAILURE) {
                return;
        }

        ZEND_FETCH_RESOURCE(f, php_fd_t *, &r_fd, -1, le_coin_acceptor_name, le_coin_acceptor);

	pit[CA_SLOT_A] = "A";
	pit[CA_SLOT_B] = "B";
	pit[CA_SLOT_C] = "C";
	pit[CA_SLOT_D] = "D";

	for (i = 0; i < 4; i++)
	{
	        switch (i)
	        {
	          case CA_SLOT_A:
	            write(f->fd, CA_PIT_A, CA_REQUEST_LENGTH);
	            break;
                  case CA_SLOT_B:
                    write(f->fd, CA_PIT_B, CA_REQUEST_LENGTH);
                    break;
                  case CA_SLOT_C:
                    write(f->fd, CA_PIT_C, CA_REQUEST_LENGTH);
                    break;
                  case CA_SLOT_D:
                    write(f->fd, CA_PIT_D, CA_REQUEST_LENGTH);
                    break;
	        }
	        
	        ca_sleep();
	        answer[i] = emalloc(CA_ANSWER_LENGTH + 1);
	        res = read(f->fd, answer[i], CA_ANSWER_LENGTH);

	        if (res <= 0)
	        {
		  efree(answer[i]);
                  php_error_docref(NULL TSRMLS_CC, E_WARNING, "no answer from device, while reading slot %s", pit[i]);
                  RETURN_FALSE;       
		}

		answer[i] = erealloc(answer[i], res + 1);
		answer[i][res] = 0;
	        
	        switch (i)
	        {
	          case CA_SLOT_A:
	            result = strncmp(answer[i], CA_PIT_A_OK, 1);
	            break;
                  case CA_SLOT_B:
                    result = strncmp(answer[i], CA_PIT_B_OK, 1);
                    break;
                  case CA_SLOT_C:
                    result = strncmp(answer[i], CA_PIT_C_OK, 1);
                    break;
                  case CA_SLOT_D:
                    result = strncmp(answer[i], CA_PIT_D_OK, 1);
                    break;
	        }
	        
	        if (result != 0)
	        {
	            php_error_docref(NULL TSRMLS_CC, E_WARNING, "device has not answered as expected, while reading slot %s", pit[i]);
	            RETURN_FALSE;       
	        }
	}

	int coins[4];
	int a;
	
	for (i = 0; i < 4; i++)
	{ 
            a = (int)answer[i][2];
            coins[i] = 0;
            
            if ((a < 0) || (a > 9))
            {
	      php_error_docref(NULL TSRMLS_CC, E_WARNING, "slot %s has stored more than 9 coins, please reset first", pit[i]);
	      RETURN_FALSE;       
            } else {
              coins[i] = a;
            }
            
            // Autoreset, if configured
            if ((COIN_ACCEPTOR_G(autoreset) == "on")
            || (COIN_ACCEPTOR_G(autoreset) == "On"))
            {
              switch (i)
              {
                case CA_SLOT_A:
                  if (coins[i] > 0) ca_reset_slot(f->fd, CA_SLOT_A);
                  break;
                case CA_SLOT_B:
                  if (coins[i] > 0) ca_reset_slot(f->fd, CA_SLOT_B);
                  break;
                case CA_SLOT_C:
                  if (coins[i] > 0) ca_reset_slot(f->fd, CA_SLOT_C);
                  break;
                case CA_SLOT_D:
                  if (coins[i] > 0) ca_reset_slot(f->fd, CA_SLOT_D);
                  break;
              }
            }
            
            efree(answer[i]);
	}
	
        // Struct return array
        array_init(return_value);
        ADD_LFIELD("slot_a", coins[CA_SLOT_A]);
        ADD_LFIELD("slot_b", coins[CA_SLOT_B]);
        ADD_LFIELD("slot_c", coins[CA_SLOT_C]);
        ADD_LFIELD("slot_d", coins[CA_SLOT_D]);
}

/* {{{ proto bool ca_reset(resource fd, int slot)
   Reset slot given by fd */
PHP_FUNCTION(ca_reset)
{
        zval     *r_fd;
        php_fd_t *f;
        int	 slot;

        if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "rl", &r_fd, &slot) == FAILURE) {
                return;
        }

        ZEND_FETCH_RESOURCE(f, php_fd_t *, &r_fd, -1, le_coin_acceptor_name, le_coin_acceptor);
        
        if (ca_reset_slot(f->fd, slot) == 0)
        {
          RETURN_TRUE;
        } else {
          RETURN_FALSE;
        }
}

/* {{{ proto void ca_close(resource fd)
   Close the descriptor given by fd */
PHP_FUNCTION(ca_close)
{
        zval     *r_fd;
        php_fd_t *f;

        if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "r", &r_fd) == FAILURE) {
                return;
        }

        ZEND_FETCH_RESOURCE(f, php_fd_t *, &r_fd, -1, le_coin_acceptor_name, le_coin_acceptor);

        zend_list_delete(Z_LVAL_P(r_fd));
}
/* }}} */
                                                                   

/*
 * Local variables:
 * tab-width: 4
 * c-basic-offset: 4
 * End:
 * vim600: noet sw=4 ts=4 fdm=marker
 * vim<600: noet sw=4 ts=4
 */
